@@ -1,4 +1,4 @@
-"""A module for Alyx file related errors
+"""ALyx file related errors
 A set of Alyx and ALF related error classes which provide a more verbose description of the raised
 issues.
 """
