@@ -1,5 +1,4 @@
-"""
-A module for inter-converting experiment identifiers.
+"""A module for inter-converting experiment identifiers.
 
 There are multiple ways to uniquely identify an experiment:
     - eid (str) : An experiment UUID as a string
@@ -14,16 +13,16 @@ import datetime
 from uuid import UUID
 from inspect import getmembers, isfunction, unwrap
 from pathlib import Path, PurePosixPath
-from typing import Optional, Union, Sequence, Mapping, List, Iterable as Iter
+from urllib.parse import urlsplit
+from typing import Optional, Union, Mapping, List, Iterable as Iter
 
 import numpy as np
 import pandas as pd
-
-import one.alf.io as alfio
 from iblutil.io import parquet
 from iblutil.util import Bunch
 
-def Listable(t): return Union[t, Sequence[t]]  # noqa
+import one.alf.io as alfio
+from .util import Listable
 
 
 def recurse(func):
@@ -40,11 +39,8 @@ def recurse(func):
 
 def parse_values(func):
     """Convert str values in reference dict to appropriate type"""
-    @functools.wraps(func)
-    def wrapper_decorator(*args, **kwargs):
-        parse = kwargs.pop('parse', True)
-        ref = func(*args, **kwargs)
-        if parse and ref:
+    def parse_ref(ref):
+        if ref:
             if isinstance(ref['date'], str):
                 if len(ref['date']) == 10:
                     ref['date'] = datetime.date.fromisoformat(ref['date'])
@@ -52,6 +48,17 @@ def parse_values(func):
                     ref['date'] = datetime.datetime.fromisoformat(ref['date']).date()
             ref['sequence'] = int(ref['sequence'])
         return ref
+
+    @functools.wraps(func)
+    def wrapper_decorator(*args, **kwargs):
+        parse = kwargs.pop('parse', True)
+        ref = func(*args, **kwargs)
+        if not parse or isinstance(ref, str):
+            return ref
+        elif isinstance(ref, list):
+            return list(map(parse_ref, ref))
+        else:
+            return parse_ref(ref)
     return wrapper_decorator
 
 
@@ -71,7 +78,7 @@ class ConversionMixin:
         if isinstance(id, UUID):
             return str(id)
         elif self.is_exp_ref(id):
-            return self.ref2eid(id, one=self)
+            return self.ref2eid(id)
         elif isinstance(id, dict):
             assert {'subject', 'number', 'start_time', 'lab'}.issubset(id)
             root = Path(cache_dir or self._cache_dir)
@@ -80,9 +87,11 @@ class ConversionMixin:
                 'Subjects', id['subject'],
                 id['start_time'][:10],
                 ('%03d' % id['number']))
-        if isinstance(id, Path) or alfio.is_session_path(id):
+        if isinstance(id, Path):
             return self.path2eid(id)
         elif isinstance(id, str):
+            if alfio.is_session_path(id) or alfio.get_session_path(id):
+                return self.path2eid(id)
             if len(id) > 36:
                 id = id[-36:]
             if not alfio.is_uuid_string(id):
@@ -113,11 +122,6 @@ class ConversionMixin:
 
         # load path from cache
         if self._index_type() is int:
-            # ids = np.array(self._cache['sessions'].index.tolist())
-            # ids = self._cache['sessions'].reset_index()[['eid_0', 'eid_1']].to_numpy()
-            # ic = find_first_2d(ids, parquet.str2np(eid))
-            # if ic is not None:
-            #     ses = self._cache['sessions'].iloc[ic]
             eid = parquet.str2np(eid).tolist()
         try:
             ses = self._cache['sessions'].loc[eid]
@@ -169,10 +173,18 @@ class ConversionMixin:
         """
         if isinstance(filepath, str) and filepath.startswith('http'):
             # Remove the UUID from path
+            filepath = urlsplit(filepath).path.strip('/')
             filepath = alfio.remove_uuid_file(PurePosixPath(filepath), dry=True)
-        session_path = '/'.join(alfio.get_session_path(filepath).parts[-5:])
-        if (rec := self._cache['datasets']).empty:
+            session_path = alfio.get_session_path(filepath).as_posix()
+        else:
+            # No way of knowing root session path parts without cache tables
+            eid = self.path2eid(filepath)
+            session_path, *_ = self.list_datasets(eid, details=True).session_path
+        rec = self._cache['datasets']
+        if rec.empty:
             return
+        # if (rec := self._cache['datasets']).empty:  # py 3.8
+        #     return
         rec = rec[rec['session_path'] == session_path]
         rec = rec[rec['rel_path'].apply(lambda x: filepath.as_posix().endswith(x))]
         assert len(rec) < 2, 'Multiple records found'
@@ -211,7 +223,7 @@ class ConversionMixin:
         :param dataset: A datasets dataframe slice
         :return: File path for the record
         """
-        assert len(dataset) == 1
+        assert isinstance(dataset, pd.Series) or len(dataset) == 1
         session_path, rel_path = dataset[['session_path', 'rel_path']].to_numpy().flatten()
         file = Path(self._cache_dir, session_path, rel_path)
         return file  # files[0] if len(datasets) == 1 else files
@@ -278,7 +290,7 @@ class ConversionMixin:
         """
         ref = self.ref2dict(ref, parse=False)  # Ensure dict
         session = self.search(
-            subjects=ref['subject'],
+            subject=ref['subject'],
             date_range=str(ref['date']),
             number=ref['sequence'])
         assert len(session) == 1, 'session not found'
@@ -304,16 +316,17 @@ class ConversionMixin:
          WindowsPath('E:/FlatIron/cortexlab/Subjects/KS005/2019-04-11/001')]
         """
         eid2path = unwrap(self.eid2path)
-        ref2eid = unwrap(self.eid2path)
-        return eid2path(ref2eid(ref))
+        ref2eid = unwrap(self.ref2eid)
+        return eid2path(self, ref2eid(self, ref))
 
     @staticmethod
     @parse_values
-    def path2ref(path_str: Union[str, Path, Iter]) -> Union[Bunch, List]:
+    def path2ref(path_str: Union[str, Path, Iter], as_dict=True) -> Union[Bunch, List]:
         """
         Returns a human readable experiment reference, given a session path.
         The path need not exist.
         :param path_str: A path to a given session
+        :param as_dict: If True a Bunch is returned, otherwise a string
         :return: one or more objects with keys ('subject', 'date', 'sequence')
 
         Examples:
@@ -331,7 +344,9 @@ class ConversionMixin:
             return [unwrap(ConversionMixin.path2ref)(x) for x in path_str]
         pattern = r'(?P<subject>[\w-]+)([\\/])(?P<date>\d{4}-\d{2}-\d{2})(\2)(?P<sequence>\d{3})'
         match = re.search(pattern, str(path_str))
-        return Bunch(match.groupdict()) if match else match
+        if match:
+            ref = match.groupdict()
+            return Bunch(ref) if as_dict else '{date:s}_{sequence:s}_{subject:s}'.format(**ref)
 
     def ref2dj(self, ref: Union[str, Mapping, Iter]):
         """
@@ -430,6 +445,8 @@ class ConversionMixin:
 
     @staticmethod
     def dict2ref(ref_dict):
+        if isinstance(ref_dict, (list, tuple)):
+            return [ConversionMixin.dict2ref(x) for x in ref_dict]
         if not ref_dict:
             return
         parsed = any(not isinstance(k, str) for k in ref_dict.values())
@@ -438,14 +455,6 @@ class ConversionMixin:
                       else '{date:s}_{sequence:s}_{subject:s}')
         return format_str.format(**ref_dict)
 
-    def to(self, eid, type):
-        if type == 'path':
-            return self.eid2path(eid)
-        elif type == 'ref':
-            return self.eid2ref(eid, as_dict=False)
-        else:
-            raise ValueError(f'Unsupported type "{type}"')
-
 
 def deprecate(func):
     """Print deprecation warning about decorated function"""
