@@ -1,4 +1,4 @@
-"""API for interacting with a remote Alyx instance
+"""API for interacting with a remote Alyx instance through REST
 The AlyxClient class contains methods for making remote Alyx REST queries and downloading remote
 files through Alyx.
 
@@ -25,6 +25,7 @@ Examples:
     local_path = self.alyx.download_file(url)
 
 TODO Move converters to another module
+TODO Add logout method to delete the active token
 """
 import json
 import logging
@@ -35,6 +36,7 @@ import functools
 import urllib.request
 from urllib.error import HTTPError
 from collections.abc import Mapping
+from typing import Optional
 from datetime import datetime, timedelta
 from pathlib import Path, PurePosixPath
 import warnings
@@ -50,40 +52,40 @@ import one.params
 from iblutil.io import hashfile
 import one.alf.io as alfio
 
-SDSC_ROOT_PATH = PurePosixPath('/mnt/ibl')
+SDSC_ROOT_PATH = PurePosixPath('/mnt/ibl')  # FIXME Move to ibllib, or get from Alyx
 _logger = logging.getLogger(__name__)
 
 
-def cache_response(func, mode='get', default_expiry=timedelta(days=1)):
+def _cache_response(method):
     """
     Decorator for the generic request method; caches the result of the query and on subsequent
     calls, returns cache instead of hitting the database
-    :param func: function to wrap (i.e. AlyxClient._generic_request)
-    :param mode: which type http method to apply cache to; if None, all requests are cached
-    :param default_expiry: The default length of time that cache file is valid for,
-    The default expiry is overridden by the `expires` kwarg.  If False, the caching is turned off.
+    :param method: function to wrap (i.e. AlyxClient._generic_request)
     :return: handle to wrapped method
     """
-    @functools.wraps(func)
-    def wrapper_decorator(*args, expires=None, clobber=False, **kwargs):
+
+    @functools.wraps(method)
+    def wrapper_decorator(alyx_client, *args, expires=None, clobber=False, **kwargs):
         """
         REST caching wrapper
+        :param alyx_client: an instance of the AlyxClient class
         :param args: positional arguments for applying to wrapped function
         :param expires: an optional timedelta for how long cached response is valid.  If True,
         the cached response will not be used on subsequent calls.  If None, the default expiry
         is applied.
         :param clobber: If True any existing cached response is overwritten
         :param kwargs: keyword arguments for applying to wrapped function
-        :return: the REST response JSON either from cached file or directly from remote endpoint
+        :return: the REST response JSON either from cached file or directly from remote
+        endpoint
         """
-        expires = expires or default_expiry
-        if mode and args[1].__name__ != mode:
-            return func(*args, **kwargs)
+        expires = expires or alyx_client.default_expiry
+        mode = (alyx_client.cache_mode or '').lower()
+        if args[0].__name__ != mode and mode != '*':
+            return method(alyx_client, *args, **kwargs)
         # Check cache
-        proc, loc = args[0].base_url.replace(':/', '').split('/')
-        rest_cache = Path(one.params.get_params_dir(), '.rest', loc, proc)
+        rest_cache = one.params.get_rest_dir(alyx_client.base_url)
         sha1 = hashlib.sha1()
-        sha1.update(bytes(args[2], 'utf-8'))
+        sha1.update(bytes(args[1], 'utf-8'))
         name = sha1.hexdigest()
         # Reversible but length may exceed 255 chars
         # name = base64.urlsafe_b64encode(args[2].encode('UTF-8')).decode('UTF-8')
@@ -96,7 +98,7 @@ def cache_response(func, mode='get', default_expiry=timedelta(days=1)):
             if datetime.fromisoformat(when) > datetime.now():
                 return cached
         try:
-            response = func(*args, **kwargs)
+            response = method(alyx_client, *args, **kwargs)
         except requests.exceptions.ConnectionError as ex:
             if cached and not clobber:
                 warnings.warn('Failed to connect, returning cached response', RuntimeWarning)
@@ -207,7 +209,7 @@ def _path_from_dataset(dset, root_path=None, repository=None, uuid=False):
     return _path_from_filerecord(fr, root_path=root_path, uuid=uuid)
 
 
-def _path_from_filerecord(fr, root_path=SDSC_ROOT_PATH, uuid=None):
+def _path_from_filerecord(fr, root_path=SDSC_ROOT_PATH, uuid=None):  # FIXME change default to '/'
     """
     Returns a data file Path constructed from an Alyx file record.  The Path type returned
     depends on the type of root_path: If root_path is a string a Path object is returned,
@@ -279,7 +281,7 @@ def http_download_file(full_link_to_file, chunks=None, *, clobber=False, silent=
 
     # default cache directory is the home dir
     if not cache_dir:
-        cache_dir = str(Path.home().joinpath("Downloads"))
+        cache_dir = str(Path.home().joinpath('Downloads'))
 
     # This is the local file name
     file_name = str(cache_dir) + os.sep + os.path.basename(full_link_to_file)
@@ -307,7 +309,7 @@ def http_download_file(full_link_to_file, chunks=None, *, clobber=False, silent=
     req = urllib.request.Request(full_link_to_file)
     if chunks is not None:
         first_byte, n_bytes = chunks
-        req.add_header("Range", "bytes=%d-%d" % (first_byte, first_byte + n_bytes - 1))
+        req.add_header('Range', 'bytes=%d-%d' % (first_byte, first_byte + n_bytes - 1))
 
     # add additional headers
     if headers is not None:
@@ -318,12 +320,12 @@ def http_download_file(full_link_to_file, chunks=None, *, clobber=False, silent=
     try:
         u = urllib.request.urlopen(req)
     except HTTPError as e:
-        _logger.error(f"{str(e)} {full_link_to_file}")
+        _logger.error(f'{str(e)} {full_link_to_file}')
         raise e
 
     file_size = int(u.getheader('Content-length'))
     if not silent:
-        print(f"Downloading: {file_name} Bytes: {file_size}")
+        print(f'Downloading: {file_name} Bytes: {file_size}')
     file_size_dl = 0
     block_sz = 8192 * 64 * 8
 
@@ -344,18 +346,16 @@ def http_download_file(full_link_to_file, chunks=None, *, clobber=False, silent=
     return (file_name, md5.hexdigest()) if return_md5 else file_name
 
 
-def file_record_to_url(file_records, urls=[]):
+def file_record_to_url(file_records):
     """
-    Translate a Json dictionary to an usable http url for downlading files.
+    Translate a Json dictionary to an usable http url for downloading files.
 
     :param file_records: json containing a 'data_url' field
     :type file_records: dict
-    :param urls: a list of strings containing previous data_urls on which new urls
-     will be appended
-    :type urls: list
 
     :return: urls: (list) a list of strings representing full data urls
     """
+    urls = []
     for fr in file_records:
         if fr['data_url'] is not None:
             urls.append(fr['data_url'])
@@ -372,18 +372,17 @@ def dataset_record_to_url(dataset_record):
     :return: (list) a list of strings representing files urls corresponding to the datasets records
     """
     urls = []
-    if type(dataset_record) is dict:
+    if isinstance(dataset_record, dict):
         dataset_record = [dataset_record]
     for ds in dataset_record:
-        urls = file_record_to_url(ds['file_records'], urls)
+        urls += file_record_to_url(ds['file_records'])
     return urls
 
 
-class UniqueSingletons(type):  # TODO Perhaps make ONE the singleton
+class UniqueSingletons(type):
     _instances: list = []
 
     def __call__(cls, *args, **kwargs):
-        # print('args', args, '\nkwargs', kwargs)
         for inst in UniqueSingletons._instances:
             if cls in inst and inst.get(cls, None).get('args') == (args, kwargs):
                 return inst[cls].get('instance')
@@ -402,52 +401,55 @@ class UniqueSingletons(type):  # TODO Perhaps make ONE the singleton
 class AlyxClient(metaclass=UniqueSingletons):
     """
     Class that implements simple GET/POST wrappers for the Alyx REST API
-    http://alyx.readthedocs.io/en/latest/api.html
+    http://alyx.readthedocs.io/en/latest/api.html  # FIXME old link
     """
     _token = None
-    _headers = None
+    _headers = None  # Headers for REST requests only
     user = None
+    base_url = None
 
-    def __init__(self, base_url=None, username=None, password=None, cache_dir=None, silent=False):
+    def __init__(self, base_url=None, username=None, password=None,
+                 cache_dir=None, silent=False, cache_rest='GET'):
         """
         Create a client instance that allows to GET and POST to the Alyx server
         For oneibl, constructor attempts to authenticate with credentials in params.py
         For standalone cases, AlyxClient(username='', password='', base_url='')
 
         :param username: Alyx database user
-        :type username: str
         :param password: Alyx database password
-        :type password: str
         :param base_url: Alyx server address, including port and protocol
-        :type base_url: str
+        :param cache_rest: which type of http method to apply cache to; if '*', all requests are
+        cached.
         """
         self.silent = silent
         self._par = one.params.get(client=base_url, silent=self.silent)
         # TODO Pass these to `get` and have it deal with setup defaults
         self._par = self._par.set('ALYX_LOGIN', username or self._par.ALYX_LOGIN)
         self._par = self._par.set('ALYX_PWD', password or self._par.ALYX_PWD)
-        self._par = self._par.set('ALYX_URL', base_url or self._par.ALYX_URL)
+        self.base_url = base_url or self._par.ALYX_URL
         self._par = self._par.set('CACHE_DIR', cache_dir or self._par.CACHE_DIR)
         self.authenticate()
         self._rest_schemes = None
         # the mixed accept application may cause errors sometimes, only necessary for the docs
         self._headers['Accept'] = 'application/json'
+        # REST cache parameters
+        # The default length of time that cache file is valid for,
+        # The default expiry is overridden by the `expires` kwarg.  If False, the caching is
+        # turned off.
+        self.default_expiry = timedelta(days=1)
+        self.cache_mode = cache_rest
         self._obj_id = id(self)
 
     @property
     def rest_schemes(self):
         """Delayed fetch of rest schemes speeds up instantiation"""
         if not self._rest_schemes:
-            self._rest_schemes = self.get('/docs')
+            self._rest_schemes = self.get('/docs', expires=timedelta(weeks=1))
         return self._rest_schemes
 
     @property
     def cache_dir(self):
-        return self._par.CACHE_DIR
-
-    @property
-    def base_url(self):
-        return self._par.ALYX_URL
+        return Path(self._par.CACHE_DIR)
 
     def is_logged_in(self):
         return self._token and self.user
@@ -460,13 +462,13 @@ class AlyxClient(metaclass=UniqueSingletons):
         EXCLUDE = ('_type', '_meta', '', 'auth-token')
         return sorted(x for x in self.rest_schemes.keys() if x not in EXCLUDE)
 
-    @cache_response
+    @_cache_response
     def _generic_request(self, reqfunction, rest_query, data=None, files=None):
         # makes sure the base url is the one from the instance
-        rest_query = rest_query.replace(self._par.ALYX_URL, '')
+        rest_query = rest_query.replace(self.base_url, '')
         if not rest_query.startswith('/'):
             rest_query = '/' + rest_query
-        _logger.debug(f"{self._par.ALYX_URL + rest_query}, headers: {self._headers}")
+        _logger.debug(f"{self.base_url + rest_query}, headers: {self._headers}")
         headers = self._headers.copy()
         if files is None:
             data = json.dumps(data) if isinstance(data, dict) or isinstance(data, list) else data
@@ -474,15 +476,16 @@ class AlyxClient(metaclass=UniqueSingletons):
         if rest_query.startswith('/docs'):
             # the mixed accept application may cause errors sometimes, only necessary for the docs
             headers['Accept'] = 'application/coreapi+json'
-        r = reqfunction(self._par.ALYX_URL + rest_query, stream=True, headers=headers,
+        r = reqfunction(self.base_url + rest_query, stream=True, headers=headers,
                         data=data, files=files)
         if r and r.status_code in (200, 201):
             return json.loads(r.text)
         elif r and r.status_code == 204:
             return
         else:
-            _logger.error(self._par.ALYX_URL + rest_query)
-            _logger.error(r.text)
+            if not self.silent:
+                _logger.error(self.base_url + rest_query)
+                _logger.error(r.text)
             raise (requests.HTTPError(r))
 
     def authenticate(self, cache_token=True, force=False):
@@ -499,10 +502,10 @@ class AlyxClient(metaclass=UniqueSingletons):
             return
         try:
             credentials = {'username': self._par.ALYX_LOGIN, 'password': self._par.ALYX_PWD}
-            rep = requests.post(self._par.ALYX_URL + '/auth-token', data=credentials)
+            rep = requests.post(self.base_url + '/auth-token', data=credentials)
         except requests.exceptions.ConnectionError:
             raise ConnectionError(
-                f"Can't connect to {self._par.ALYX_URL}.\n" +
+                f"Can't connect to {self.base_url}.\n" +
                 "IP addresses are filtered on IBL database servers. \n" +
                 "Are you connecting from an IBL participating institution ?"
             )
@@ -519,7 +522,7 @@ class AlyxClient(metaclass=UniqueSingletons):
             one.params.save(par, self.base_url)
         self.user = self._par.ALYX_LOGIN
         if not self.silent:
-            print(f"Connected to {self._par.ALYX_URL} as {self._par.ALYX_LOGIN}")
+            print(f"Connected to {self.base_url} as {self.user}")
 
     def delete(self, rest_query):
         """
@@ -549,7 +552,6 @@ class AlyxClient(metaclass=UniqueSingletons):
             url = (self._validate_file_url(x) for x in url)
             download_fcn = http_download_file_list
         pars = dict(
-            headers=self._headers,
             silent=kwargs.pop('silent', self.silent),
             cache_dir=kwargs.pop('cache_dir', self._par.CACHE_DIR),
             username=self._par.HTTP_DATA_SERVER_LOGIN,
@@ -564,6 +566,7 @@ class AlyxClient(metaclass=UniqueSingletons):
         :return: List of parquet table file paths
         """
         # query the database for the latest cache; expires=None overrides cached response
+        self.cache_dir.mkdir(exist_ok=True)
         with tempfile.TemporaryDirectory(dir=self.cache_dir) as tmp:
             file = http_download_file(f'{self.base_url}/cache.zip',
                                       username=self._par.ALYX_LOGIN,
@@ -677,16 +680,16 @@ class AlyxClient(metaclass=UniqueSingletons):
         alyx_client.rest(endpoint): lists actions for endpoint
         alyx_client.rest(endpoint, action): lists fields and URL
 
-        Example with a rest endpoint with all actions
+        Example REST endpoint with all actions:
 
-        >>> alyx.client.rest('subjects', 'list')
-            alyx.client.rest('subjects', 'list', field_filter1='filterval')
-            alyx.client.rest('subjects', 'create', data=sub_dict)
-            alyx.client.rest('subjects', 'read', id='nickname')
-            alyx.client.rest('subjects', 'update', id='nickname', data=sub_dict)
-            alyx.client.rest('subjects', 'partial_update', id='nickname', data=sub_dict)
-            alyx.client.rest('subjects', 'delete', id='nickname')
-            alyx.client.rest('notes', 'create', data=nd, files={'image': open(image_file, 'rb')})
+            client.rest('subjects', 'list')
+            client.rest('subjects', 'list', field_filter1='filterval')
+            client.rest('subjects', 'create', data=sub_dict)
+            client.rest('subjects', 'read', id='nickname')
+            client.rest('subjects', 'update', id='nickname', data=sub_dict)
+            client.rest('subjects', 'partial_update', id='nickname', data=sub_dict)
+            client.rest('subjects', 'delete', id='nickname')
+            client.rest('notes', 'create', data=nd, files={'image': open(image_file, 'rb')})
 
         :param url: endpoint name
         :param action: 'list', 'create', 'read', 'update', 'partial_update', 'delete'
@@ -721,7 +724,7 @@ class AlyxClient(metaclass=UniqueSingletons):
             pprint(list(endpoint_scheme.keys()))
             return
         # make sure the the desired action exists, if not throw an informative error
-        if action not in endpoint_scheme.keys():
+        if action not in endpoint_scheme:
             raise ValueError('Action "' + action + '" for REST endpoint "' + endpoint + '" does ' +
                              'not exist. Available actions are: ' +
                              '\n       ' + '\n       '.join(endpoint_scheme.keys()))
@@ -870,7 +873,7 @@ class AlyxClient(metaclass=UniqueSingletons):
             uuid: str = None,
             field_name: str = 'json',
             key: str = None
-    ) -> dict:
+    ) -> Optional[dict]:
         """json_field_remove_key
         Will remove inputted key from json field dict and reupload it to Alyx.
         Needs endpoint, uuid and json field name
@@ -894,7 +897,7 @@ class AlyxClient(metaclass=UniqueSingletons):
         # if contents are not dict, cannot remove key, return contents
         if isinstance(current, str):
             _logger.warning(f"Cannot remove key {key} content of json field is of type str")
-            return current
+            return None
         # If key not present in contents of json field cannot remove key, return contents
         if current.get(key, None) is None:
             _logger.warning(
@@ -917,6 +920,5 @@ class AlyxClient(metaclass=UniqueSingletons):
         return _[field_name]
 
     def clear_rest_cache(self):
-        proc, loc = self.base_url.replace(':/', '').split('/')
-        for file in Path(one.params.get_params_dir(), '.rest', loc, proc).glob('*'):
+        for file in one.params.get_rest_dir(self.base_url).glob('*'):
             file.unlink()
